# Programmed with <3 by fluffy

from pygame import Rect
from pygame import Surface
from . import constants
from .color import Color
from .event import Event
from .font import Font
from .hitbox import Hitbox
from .hitbox import HitboxManager
from .hitbox import SubHitbox
from .image import Image
from .mouse import Mouse
from .object import Draggable
from .object import Eventable
from .object import Object
from .object import ObjectHolder
from .object import ObjectManager
from .object import Renderable
from .object import Updateable
from .position import Position
from .position import PositionReference
from .render_context import RenderContext
from .window import Window


# Programmed with <3 by fluffy

from pygame import Rect
from dataclasses import dataclass
from . import constants
from .event import Event
from .hitbox import Hitbox
from .hitbox import HitboxManager
from .position import Position
from .render_context import RenderContext

class Object:
    def __init__(self, position:Position, hitbox:Hitbox=None) -> None:
        self.position = position
        if hitbox is not None: hitbox.bind_master(self)
        self.__hitbox = hitbox
    
    def is_colliding(self) -> bool:
        if self.__hitbox is None: return 0
        return self.__hitbox.is_active()

    def has_hitbox(self) -> bool:
        return not self.__hitbox is None

    def get_hitbox(self) -> Hitbox:
        return self.__hitbox

class Renderable():
    def on_render(self, ctx:RenderContext) -> None: pass

class Updateable():
    def on_update(self, dt:float) -> None: pass

class Eventable():
    def init_eventable(self, events:list[Event], only_on_active_hitbox:bool=1) -> None:
        # NOTE THIS VARIABLES ARE USED IN THE OBJECTMANAGER CLASS
        self.__only_on_active_hitbox = only_on_active_hitbox
        self.__events = list(set(events))

    def on_event(self, event:Event) -> None: pass

class Draggable(Object, Eventable):
    def __init__(self, position:Position, hitbox:Hitbox=None) -> None:
        super().__init__(position, hitbox)

    def init_draggable(self, events:list[Event], drag_buttons:list[int]=None, boundry_box:Rect=None) -> None:
        super().init_eventable([constants.MOUSEBUTTONDOWN, constants.MOUSEBUTTONUP, constants.MOUSEMOTION, *events], 0)
        self.__dragging_boundry_box = boundry_box
        self.__drag_buttons = [constants.BUTTON_LEFT] if drag_buttons is None else drag_buttons
        self.__draggable_active: bool = 0
        self.__drag_locked_x = 0
        self.__drag_locked_y = 0
        self.__drag_start_pos = None
    
    def object_registered(self, object_manager:'ObjectManager') -> None:
        hitbox_manager = object_manager.get_hitbox_manager()
        if not hitbox_manager.has_standard_boundry_box(): return
        if self.__dragging_boundry_box is constants.DEFAULT_BOUNDRY: self.__dragging_boundry_box = hitbox_manager.get_standard_boundry_box()

    def get_drag_start_pos(self) -> Position:
        return self.__drag_start_pos

    def set_locked(self, x:bool=1, y:bool=1) -> None:
        self.__drag_locked_x = x
        self.__drag_locked_y = y

    def on_event(self, event:Event) -> None:
        if self.__drag_locked_x and self.__drag_locked_y: return
        if not self.has_hitbox(): return
        match event.type:
            case constants.MOUSEBUTTONDOWN:
                if self.__draggable_active: return
                if not event.button in self.__drag_buttons: return
                if not self.get_hitbox().is_active(): return
                self.__drag_start_pos = self.position.clone()
                self.__draggable_active = 1
            case constants.MOUSEBUTTONUP:
                if not self.__draggable_active: return
                self.__draggable_active = 0
                self.__drag_start_pos = None
            case constants.MOUSEMOTION:
                if not self.__draggable_active: return
                relx, rely = event.rel
                if self.__drag_locked_x: relx = 0
                if self.__drag_locked_y: rely = 0
                self.position.move_rel(relx, rely)
                if not self.__dragging_boundry_box is None:
                    if self.position.x <= self.__dragging_boundry_box.x:
                        self.position.x = self.__dragging_boundry_box.x
                    elif self.position.x >= self.__dragging_boundry_box.x+self.__dragging_boundry_box.w-self.get_hitbox().w:
                        self.position.x = self.__dragging_boundry_box.x+self.__dragging_boundry_box.w-self.get_hitbox().w
                    if self.position.y <= self.__dragging_boundry_box.y:
                        self.position.y = self.__dragging_boundry_box.y
                    elif self.position.y >= self.__dragging_boundry_box.y+self.__dragging_boundry_box.h-self.get_hitbox().h:
                        self.position.y = self.__dragging_boundry_box.y+self.__dragging_boundry_box.h-self.get_hitbox().h

@dataclass
class ObjectHolder:
    obj:Object
    object_id:int
    renderable:bool
    updateable:bool
    eventable:bool

class ObjectManager:
    def __init__(self, hitbox_manager:HitboxManager, use_recycled_ids:bool=1, use_inbuild_z_system:bool=1) -> None:
        self.__hitbox_manager = hitbox_manager
        self.__objects: dict[int, ObjectHolder] = {}
        self.__renderables: dict[int, ObjectHolder] = {}
        self.__updateables: dict[int, ObjectHolder] = {}
        self.__eventcallables: dict[int, ObjectHolder] = {}
        self.__object_id: int = 0
        if use_recycled_ids: self.__recycled_ids: list[int] = []
        self.__use_recycled_ids = use_recycled_ids
        self.__use_inbuild_z_system = use_inbuild_z_system
        if self.__use_inbuild_z_system: self.__current_z = 0

    def get_object_holder(self, object_id:int) -> ObjectHolder:
        if not object_id in self.__objects: return None
        return self.__objects[object_id]

    def get_hitbox_manager(self) -> HitboxManager:
        return self.__hitbox_manager

    def register(self, obj:Object) -> ObjectHolder:
        if self.__use_recycled_ids and len(self.__recycled_ids) > 0:
            new_id: int = self.__recycled_ids[0]
            del self.__recycled_ids[0]
        else:
            new_id: int = self.__object_id
            self.__object_id += 1
        holder = ObjectHolder(
            obj,
            new_id,
            isinstance(obj, Renderable),
            isinstance(obj, Updateable),
            isinstance(obj, Eventable)
        )
        self.__objects[new_id] = holder
        if holder.renderable: self.__renderables[new_id] = holder
        if holder.updateable: self.__updateables[new_id] = holder
        if holder.eventable: self.__eventcallables[new_id] = holder
        if obj.has_hitbox():
            self.__hitbox_manager.register(obj.get_hitbox())
        if self.__use_inbuild_z_system:
            obj.position.z = self.__current_z
            self.__current_z += 1
        if hasattr(obj, 'object_registered'): obj.object_registered(self)
        return holder

    def unregister(self, object_id:int) -> bool:
        if not object_id in self.__objects: return 0
        removed_obj_holder: ObjectHolder = self.__objects[object_id]
        del self.__objects[object_id]
        if removed_obj_holder.renderable: del self.__renderables[object_id]
        if removed_obj_holder.updateable: del self.__updateables[object_id]
        if removed_obj_holder.eventable: del self.__eventcallables[object_id]
        if removed_obj_holder.obj.has_hitbox(): self.__hitbox_manager.unregister(removed_obj_holder.obj.get_hitbox().get_hitbox_manager_id())
        if self.__use_recycled_ids: self.__recycled_ids.append(object_id)
        if hasattr(removed_obj_holder.obj, 'object_unregistered'): removed_obj_holder.obj.object_unregistered(self)
        return 1

    def eventcall_all(self, event:Event) -> None:
        for holder in self.__eventcallables.values():
            obj: Eventable = holder.obj
            if not event.type in obj._Eventable__events: continue
            if obj._Eventable__only_on_active_hitbox:
                if not holder.obj.has_hitbox(): continue
                if not holder.obj.get_hitbox().is_active(): continue
            obj.on_event(event)

    def update_all(self, dt:float) -> None:
        for holder in self.__updateables.values():
            obj: Updateable = holder.obj
            obj.on_update(dt)

    def render_all(self, ctx:RenderContext) -> None:
        holders: list[ObjectHolder] = [*self.__renderables.values()]
        holders.sort(key=lambda holder: holder.obj.position.z)
        for holder in self.__renderables.values():
            obj: Renderable = holder.obj
            obj.on_render(ctx)


# Programmed with <3 by fluffy

from dataclasses import dataclass

@dataclass
class Event:
    type:int
    dict:dict

    def __post_init__(self) -> None:
        for key, value in self.dict.items():
            setattr(self, key, value)


# Programmed with <3 by fluffy

import pygame

from dataclasses import dataclass

@dataclass
class Image:
    path:str

    def load(self) -> 'Image':
        self.__loaded = pygame.image.load(self.path)
        return self

    def get_loaded(self) -> pygame.Surface:
        return self.__loaded


# Programmed with <3 by fluffy

from typing import Any
import pygame

from . import Rect
from .color import Color
from .event import Event
from .hitbox import HitboxManager
from .image import Image
from .mouse import Mouse
from .object import ObjectManager
from .render_context import RenderContext

class Window:
    def __init__(self, width:int, height:int, title:str, icon:Image=None, fps:int=120, default_color:Color=None) -> None:
        pygame.init()
        self.set_title(title)
        if not icon is None: self.set_icon(icon)
        self.__screen = pygame.display.set_mode((width, height))
        self.__width, self.__height = width, height
        self.__clock = pygame.time.Clock()
        self.set_fps(fps)

        self.__render_context = RenderContext(self.__screen)
        self.__default_color = Color(0, 0, 0) if default_color is None else default_color
        self.__mouse = Mouse()

        self.__running = 0
        self.__functions = {}

        self.__setups = []

    def setup_hitbox_manager(self, use_recycled_ids:bool=1) -> HitboxManager:
        self.__hitbox_manager = HitboxManager(use_recycled_ids, Rect(0, 0, self.__width, self.__height))
        self.__setups.append('HITBOX_MANAGER')
        return self.__hitbox_manager
    
    def get_hitbox_manager(self) -> HitboxManager:
        if not 'HITBOX_MANAGER' in self.__setups: raise ValueError('Hitboxmanager is not setup')
        return self.__hitbox_manager

    def setup_object_manager(self, use_recycled_ids:bool=1, use_inbuild_z_system:bool=1) -> ObjectManager:
        if not 'HITBOX_MANAGER' in self.__setups: raise ValueError('Hitboxmanager is not setup')
        self.__object_manager = ObjectManager(self.__hitbox_manager, use_recycled_ids, use_inbuild_z_system)
        self.__setups.append('OBJECT_MANAGER')
        return self.__object_manager
    
    def get_object_manager(self) -> ObjectManager:
        if not 'OBJECT_MANAGER' in self.__setups: raise ValueError('Objectmanager is not setup')
        return self.__object_manager

    def get_mouse(self) -> Mouse:
        return self.__mouse

    def set_icon(self, image:Image) -> None:
        pygame.display.set_icon(image.get_loaded())

    def set_title(self, text:str) -> None:
        pygame.display.set_caption(text)

    def set_fps(self, fps:int) -> None:
        self.__fps = fps
        self.__fpms = fps / 10

    def stop(self) -> None:
        self.__running = 0

    def on(self, function:callable) -> None:
        self.on_raw(function.__name__, function)

    def on_raw(self, name:str, function:callable) -> None:
        self.__functions[name] = function

    def __call(self, name:str, args:list[Any]) -> None:
        if not name in self.__functions: return
        self.__functions[name](*args)

    def mainloop(self) -> None:
        deltatime = 0
        self.__running = 1
        while 1:
            for event in pygame.event.get():
                if event.type == pygame.QUIT: self.stop()
                else: self.__call('on_event', [Event(event.type, event.dict)])
            if not self.__running: break
            self.__call('on_update', [deltatime])
            self.__render_context.fill(self.__default_color)
            self.__call('on_render', [self.__render_context])
            self.__mouse.call_mouse_override(self.__render_context)
            pygame.display.flip()
            deltatime = self.__clock.tick(self.__fps) / self.__fpms


# Programmed with <3 by fluffy

from dataclasses import dataclass

@dataclass
class Color:
    r:int
    g:int
    b:int
    a:int=255

    def pack(self) -> tuple[int, int, int, int]:
        return self.r, self.g, self.b, self.a


# Programmed with <3 by fluffy

import pygame
from . import Rect
from . import Surface
from .color import Color
from .font import Font

class RenderContext:
    def __init__(self, surface:Surface) -> None:
        self.__surface:pygame.Surface = surface

    def fill(self, color:Color) -> None:
        self.__surface.fill(color.pack())

    def line(self, color:Color, x_start:int, y_start:int, x_end:int, y_end:int, width:int=1) -> None:
        pygame.draw.line(self.__surface, color.pack(), (x_start, y_start), (x_end, y_end), width)

    def rect(self, color:Color, x:int, y:int, w:int, h:int, width:int=1) -> None:
        pygame.draw.rect(self.__surface, color.pack(), (x, y, w, h), width)

    def register_font(self, font_name:str, font_size:int, bold:bool=0, italic:bool=0) -> Font:
        return Font(font_name, font_size, pygame.font.SysFont(font_name, font_size, bold, italic), bold, italic)

    def blit(self, surface:Surface, x:int=0, y:int=0, area:Rect=None) -> None:
        self.__surface.blit(surface, (x, y), area)


# Programmed with <3 by fluffy

from dataclasses import dataclass

@dataclass
class Position:
    x:int
    y:int
    z:int=0

    def pack(self, with_z:bool=0) -> tuple[int]:
        if with_z: return self.x, self.y, self.z
        return self.x, self.y

    def move_rel(self, relx:int=0, rely:int=0, relz:int=0) -> None:
        self.x += relx
        self.y += rely
        self.z += relz

    def move_to(self, x:int=None, y:int=None, z:int=None) -> None:
        if not x is None: self.x = x
        if not y is None: self.y = y
        if not z is None: self.z = z

    def clone(self) -> 'Position':
        return Position(*self.pack())
    
    def gx(self) -> int:
        return self.x
    
    def gy(self) -> int:
        return self.y

@dataclass
class PositionReference:
    position:Position
    relx:int=0
    rely:int=0

    def gx(self) -> int:
        return self.position.x+self.relx

    def gy(self) -> int:
        return self.position.y+self.rely
    
    def pack(self) -> tuple[int]:
        return self.position.x+self.relx, self.position.y+self.rely


# Programmed with <3 by fluffy

from pygame.font import Font as PyFont
from pygame import Surface
from .color import Color

class Font:
    def __init__(self, name:str, size:int, font_data:PyFont, bold:bool=0, italic:bool=0) -> None:
        self.__name:str = name
        self.__size:int = size
        self.__font_data:PyFont = font_data
        self.__bold:bool = bold
        self.__italic:bool = italic
        self.__last_rendered:Surface = None
    
    def get_name(self) -> str:
        return self.__name
    
    def get_size(self) -> int:
        return self.__size

    def is_bold(self) -> bool:
        return self.__bold

    def is_italic(self) -> bool:
        return self.__italic
    
    def render(self, text:str, color:Color, antialias:bool=1, background_color:Color=None) -> Surface:
        self.__last_rendered = self.__font_data.render(text, antialias, color.pack(), None if background_color is None else background_color.pack())
        return self.__last_rendered

    def get_last_rendered(self) -> Surface:
        return self.__last_rendered


# Programmed with <3 by fluffy

from typing import Callable
from .render_context import RenderContext
import pygame

class Mouse:
    def __init__(self) -> None:
        self.__mouse_override:Callable = None

    def get_pos(self) -> tuple:
        return pygame.mouse.get_pos()

    def set_pos(self, x:int, y:int) -> None:
        pygame.mouse.set_pos((x, y))

    def show(self) -> None:
        pygame.mouse.set_visible(1)

    def hide(self) -> None:
        pygame.mouse.set_visible(0)
    
    def set_mouse_override(self, function:Callable) -> None:
        self.__mouse_override = function

    def has_mouse_override(self) -> bool:
        return not self.__mouse_override is None

    def call_mouse_override(self, ctx:RenderContext) -> None:
        if self.has_mouse_override: self.__mouse_override(ctx, self)


# Programmed with <3 by fluffy

from pygame import Rect
from dataclasses import dataclass
from .color import Color
from .position import Position
from .render_context import RenderContext
from typing import Any

@dataclass
class SubHitbox:
    relx:int
    rely:int
    w:int
    h:int

    def is_hovering(self, position:Position, mx:int, my:int) -> bool:
        return Rect(position.x+self.relx+position.relx, position.y+self.rely+position.rely, self.w, self.h).collidepoint(mx, my)

    def render(self, ctx:RenderContext, position:Position, active:bool) -> None:
        ctx.rect(Color(0, 255, 0) if active else Color(255, 255, 255), position.gx()+self.relx, position.gy()+self.rely, self.w, self.h)

@dataclass
class Hitbox:
    position:Position
    w:int
    h:int
    subs:list[SubHitbox] = None
    relx:int=0
    rely:int=0

    def __post_init__(self) -> None:
        self.__is_active: bool = 0
        if self.subs is None: self.subs: list[SubHitbox] = []
        self.__hitbox_manager_id: int = None
        self.__master = None

    def bind_master(self, master:Any) -> None:
        self.__master = master
    
    def get_master(self) -> Any:
        return self.__master

    def _set_hitbox_manager_id(self, hitbox_id:int) -> None:
        self.__hitbox_manager_id = hitbox_id
    
    def get_hitbox_manager_id(self) -> int | None:
        return self.__hitbox_manager_id
    
    def set_active(self, state:bool) -> None:
        self.__is_active = state

    def is_active(self) -> bool:
        return self.__is_active
    
    def is_hovering(self, mx:int, my:int) -> bool:
        for sub in self.subs:
            if sub.is_hovering(self.position, mx, my): return 1
        return Rect(self.position.gx(), self.position.gy(), self.w, self.h).collidepoint(mx, my)

    def render(self, ctx:RenderContext) -> None:
        for sub in self.subs: sub.render(ctx, self.position, self.__is_active)
        ctx.rect(Color(255, 0, 0) if self.__is_active else Color(255, 255, 255), *self.position.pack(), self.w, self.h)

class HitboxManager:
    def __init__(self, use_recycled_ids:bool=1, standard_boundry_box:Rect=None) -> None:
        self.__hitboxes:dict[int, Hitbox] = {}
        self.__hitbox_id: int = 0
        if use_recycled_ids: self.__recycled_ids: list[int] = []
        self.__use_recycled_ids: bool = use_recycled_ids
        self.__standard_boundry_box = standard_boundry_box
        self.__active: Hitbox = None

    def has_standard_boundry_box(self) -> bool:
        return not self.__standard_boundry_box is None

    def get_standard_boundry_box(self) -> Rect:
        return self.__standard_boundry_box

    def register(self, hitbox:Hitbox, bind_master:bool=0) -> int:
        if self.__use_recycled_ids and len(self.__recycled_ids) > 0:
            new_id: int = self.__recycled_ids[0]
            del self.__recycled_ids[0]
        else:
            new_id: int = self.__hitbox_id
            self.__hitbox_id += 1
        self.__hitboxes[new_id] = hitbox
        if bind_master: hitbox.bind_master(self)
        hitbox._set_hitbox_manager_id(new_id)
        return new_id

    def unregister(self, hitbox_id:int) -> bool:
        if not hitbox_id in self.__hitboxes: return 0
        del self.__hitboxes[hitbox_id]
        if self.__use_recycled_ids: self.__recycled_ids.append(hitbox_id)
        return 1

    def render_all(self, ctx:RenderContext) -> None:
        hitboxes = [*self.__hitboxes.values()]
        hitboxes.sort(key=lambda hitbox: hitbox.position.z)
        for hitbox in hitboxes:
            hitbox.render(ctx)

    def update_hovering(self, mx:int, my:int) -> None:
        active:list[Hitbox] = [hitbox for hitbox in self.__hitboxes.values() if hitbox.is_hovering(mx, my)]
        if len(active) == 0:
            if not self.__active is None:
                self.__active.set_active(0)
                self.__active = None
            return
        if len(active) > 1: active.sort(key=lambda hitbox: hitbox.position.z, reverse=1)
        if self.__active != active[0]:
            active[0].set_active(1)
            if not self.__active is None: self.__active.set_active(0)
            self.__active = active[0]
    
    def get_active(self) -> Hitbox:
        return self.__active